{ File    : $URL: file:///D:/SVN/svnDelphi/trunk/Geheimakte%20Tunguska/MakSPR/src/MakSPR.dpr $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2011-10-16 23:40:29 +0200 (So, 16 Okt 2011) $
  Revision: $Rev: 35 $

  Animation Arts SPR creator.
}
program MakSPR;

{$APPTYPE CONSOLE}

{$R '..\res\version.res' '..\res\version.rc'}

uses
  KeyCfg in 'backend\KeyCfg.pas',
  CompSLZX in 'compress\CompSLZX.pas',
  CompSSHC in 'compress\CompSSHC.pas',
  Shared in 'shared\Shared.pas',
  SCRC in 'utils\SCRC.pas',
  Heap2 in 'utils\Heap2.pas',
  MemSearch in 'utils\MemSearch.pas',
  SysUtils,
  Windows;

const
  ExtToCompress: array[0..4] of AnsiString =
    ('.csv', '.dlg', '.ini', '.sdf', '.wpd');
  ExtToCompressSave: array[0..0] of AnsiString =
    ('.dat');

  CompMethod_None: array[0..3] of AnsiChar = '';
  CompMethod_Auto: array[0..3] of AnsiChar = '????';

type
  TParamRec = record
    FileSpec  : AnsiString;
    DirSpec   : AnsiString;
    KeyDomain : Longword;
    IsDemo    : Boolean;
    IsSave    : Boolean;
    NoCompress: Boolean;
  end;

  TSPRBlockDesc = packed record
    Num   : Longint;
    Offset: Longint;
    Size  : Longint
  end;

  TSPRHeader_Release = packed record
    ID         : array[0..3] of Char;
    Unknown1   : Longword;
    Seed       : Longword;
    Unknown2   : Longword;
    Unknown3   : Longword;
    Strings    : TSPRBlockDesc;
    Files      : TSPRBlockDesc;
    Directories: TSPRBlockDesc;
  end;

  TSPRHeader_Demo = packed record
    ID         : array[0..3] of Char;
    Unknown1   : Longword;
    Unknown2   : Longword;
    Seed       : Longword;
    Unknown3   : Longword;
    Unknown4   : Longword;
    Strings    : TSPRBlockDesc;
    Files      : TSPRBlockDesc;
    Unknown5   : Longword;
    Directories: TSPRBlockDesc;
  end;

  // unified header variant
  TSPRHeader = record
    Seed       : Longword;
    Strings    : TSPRBlockDesc;
    Files      : TSPRBlockDesc;
    Directories: TSPRBlockDesc;
  end;

  TSPRDate = TSystemTime;
  { format:
  TSPRDate = packed record
    Year     : Word;
    Month    : Word;
    DayOfWeek: Word;
    Day      : Word;
    Hour     : Word;
    Minute   : Word;
    Second   : Word;
    MSecond  : Word;
  end;}

  TFileEntry = packed record
    DirIdx    : Longint;
    NameIdx   : Longint;
    SuffixIdx : Longint;
    CRC32     : Longword;
    DateStamp : TSPRDate;
    Size      : Longint;
    HeadOffset: Longint;
    HeadSize  : Longint;
    DataOffset: Longint;
    DataSize  : Longint;
  end;

  TDirEntry = packed record
    NameIdx  : Longint;
    ParentIdx: Longint;
  end;

  TNameTable = array of string;
  TFileTable = packed array of TFileEntry;
  TDirTable  = packed array of TDirEntry;

var
  ParamRec   : TParamRec;

function OpenInFile(var InFile: file; const InName: AnsiString): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(InFile, InName);

  {$I-}
  Reset(InFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function OpenSPRFile(var SPRFile: file; const SPRName: AnsiString): Boolean;
begin
  FileMode := fmOpenWrite;
  AssignFile(SPRFile, SPRName);

  {$I-}
  Rewrite(SPRFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

procedure EncryptBuffer(const Buffer: TBuffer; const SPRHeader: TSPRHeader);
var
  i        : Longint;
  KeyDomain,
  Key      : Longword;
  CurByte  : Byte;
begin
  KeyDomain := ParamRec.KeyDomain * SPRHeader.Seed;
  Key := ((SPRHeader.Seed shl 8) or ((not SPRHeader.Seed) and $000000ff));

  for i := 0 to High(Buffer) do
  begin
    Key := (Key * Key) mod KeyDomain;
    CurByte := Buffer[i];
    Buffer[i] := CurByte xor Key;
    Key := Key + CurByte + 1
  end
end;

function GetDateStamp(const FileName: string): TSystemTime;
var
  CreationTime,
  LastAccessTime,
  LastWriteTime : TFileTime;
  Handle        : Longword;
begin
  FillChar(Result, SizeOf(Result), 0);

  Handle := FileOpen(FileName, fmOpenRead);
  if Handle <> INVALID_HANDLE_VALUE then
  begin
    if GetFileTime(Handle, @CreationTime, @LastAccessTime, @LastWriteTime) then
    begin
      FileTimeToLocalFileTime(LastWriteTime, LastWriteTime);
      FileTimeToSystemTime(LastWriteTime, Result)
    end;

    FileClose(Handle)
  end
end;

function WriteHeader(var SPRFile: file; const SPRHeader: TSPRHeader): Boolean;
var
  SPRRelease  : TSPRHeader_Release;
  SPRDemo     : TSPRHeader_Demo;
  BytesWritten: Longint;
begin
  Result := False;

  {$I-}
  Seek(SPRFile, 0);
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('error seeking to beginning of output');
    Exit
  end;

  if ParamRec.IsDemo then
  begin
    with SPRDemo do
    begin
      ID          := 'SPCD';
      Unknown1    := 0;
      Unknown2    := 0;
      Seed        := SPRHeader.Seed;
      Unknown3    := 0;
      Unknown4    := 0;
      Strings     := SPRHeader.Strings;
      Files       := SPRHeader.Files;
      Unknown5    := 0;
      Directories := SPRHeader.Directories
    end;

    BlockWrite(SPRFile, SPRDemo, SizeOf(SPRDemo), BytesWritten);
    if BytesWritten <> SizeOf(SPRDemo) then
    begin
      WriteLn('error writing header');
      Exit
    end
  end
  else
  begin
    with SPRRelease do
    begin
      ID          := 'SPCR';
      Unknown1    := 0;
      Seed        := SPRHeader.Seed;
      Unknown2    := 0;
      Unknown3    := 0;
      Strings     := SPRHeader.Strings;
      Files       := SPRHeader.Files;
      Directories := SPRHeader.Directories
    end;

    BlockWrite(SPRFile, SPRRelease, SizeOf(SPRRelease), BytesWritten);
    if BytesWritten <> SizeOf(SPRRelease) then
    begin
      WriteLn('error writing header');
      Exit
    end
  end;

  Result := True
end;

function WriteNameTable(var SPRFile: file; var SPRHeader: TSPRHeader; const
  NameTable: TNameTable): Boolean;
var
  i,
  Len,
  BufIdx: Longint;
  Size  : Word;
  Buffer: TBuffer;
begin
  // convert name table to buffer structure
  Len := 0;
  for i := 0 to High(NameTable) do
    Inc(Len, Word(Length(NameTable[i])) + 2);
  SetLength(Buffer, Len);

  BufIdx := Length(NameTable) * 2;
  for i := 0 to High(NameTable) do
  begin
    Size := Length(NameTable[i]);
    Move(Size, Buffer[i * 2], 2);
    Move(NameTable[i][1], Buffer[BufIdx], Size);
    Inc(BufIdx, Size)
  end;

  // encrypt name table
  EncryptBuffer(Buffer, SPRHeader);

  // store name table metadata in header
  with SPRHeader.Strings do
  begin
    Num    := Length(NameTable);
    Offset := FilePos(SPRFile);
    Size   := Length(Buffer)
  end;

  // store name table
  {$I-}
  BlockWrite(SPRFile, Buffer[0], Length(Buffer));
  {$I+}
  Result := (IOResult = 0);
  if not Result then
    WriteLn('error writing name table');

  SetLength(Buffer, 0)
end;

function WriteDirTable(var SPRFile: file; var SPRHeader: TSPRHeader;
  const DirTable: TDirTable): Boolean;
var
  Buffer: TBuffer;
begin
  // copy directory table to buffer structure
  SetLength(Buffer, Length(DirTable) * SizeOf(TDirEntry));
  Move(DirTable[0], Buffer[0], Length(Buffer));

  // encrypt directory table
  EncryptBuffer(Buffer, SPRHeader);

  // store directory table metadata in header
  with SPRHeader.Directories do
  begin
    Num := Length(DirTable);
    Offset := FilePos(SPRFile);
    Size := Length(Buffer)
  end;

  // store directory table
  {$I-}
  BlockWrite(SPRFile, Buffer[0], Length(Buffer));
  {$I+}
  Result := (IOResult = 0);
  if not Result then
    WriteLn('error writing directory table');

  SetLength(Buffer, 0)
end;

function WriteFileTable(var SPRFile: file; var SPRHeader: TSPRHeader;
  const FileTable: TFileTable): Boolean;
var
  Buffer: TBuffer;
begin
  // copy file table to buffer structure
  SetLength(Buffer, Length(FileTable) * SizeOf(TFileEntry));
  Move(FileTable[0], Buffer[0], Length(Buffer));

  // encrypt directory table
  EncryptBuffer(Buffer, SPRHeader);

  // store directory table metadata in header
  with SPRHeader.Files do
  begin
    Num := Length(FileTable);
    Offset := FilePos(SPRFile);
    Size := Length(Buffer)
  end;

  // store directory table
  {$I-}
  BlockWrite(SPRFile, Buffer[0], Length(Buffer));
  {$I+}
  Result := (IOResult = 0);
  if not Result then
    WriteLn('error writing file table');

  SetLength(Buffer, 0)
end;

function AddToNameTable(const Name: AnsiString;
  var NameTable: TNameTable): Longint;
var
  i: Longint;
begin
  Result := -1;

  for i := 0 to High(NameTable) do
    if NameTable[i] = Name then
      Result := i;

  if Result = -1 then
  begin
    SetLength(NameTable, Length(NameTable) + 1);

    Result := High(NameTable);
    NameTable[Result] := Name
  end
end;

function Compress(var Buffer: TBuffer; var FileEntry: TFileEntry;
  const NameTable: TNameTable): Longint;
var
  i,
  ResultSLZX,
  ResultSSHC    : Longint;
  CompMethod    : AnsiString;
  CompSuccess   : Boolean;
  CompBufferSLZX,
  CompBufferSSHC: TBuffer;
begin
  Result := 0;
  ResultSSHC := 0;
  ResultSLZX := 0;

  // determine compression method
  CompMethod := CompMethod_None;

  if ParamRec.NoCompress then
    Exit;

  if ParamRec.IsSave then
  begin
    // in savegames, all text files are SSHC-compressed
    for i := 0 to High(ExtToCompressSave) do
      if ExtToCompressSave[i] = NameTable[FileEntry.SuffixIdx] then
      begin
        CompMethod := CompMethod_SSHC;
        Break
      end
  end
  else
    // in archives, text files can use either SLZX or SSHC compression
    for i := 0 to High(ExtToCompress) do
      if ExtToCompress[i] = NameTable[FileEntry.SuffixIdx] then
      begin
        CompMethod := CompMethod_Auto;
        Break
      end;

  if CompMethod = CompMethod_None then
    Exit;

  if CompMethod = CompMethod_Auto then
  begin
    ResultSLZX := Compress_SLZX(Buffer, CompBufferSLZX, CompSuccess);
    ResultSSHC := Compress_SSHC(Buffer, CompBufferSSHC);

    if (Length(CompBufferSLZX) > Length(Buffer)) or not CompSuccess or
      (Length(CompBufferSLZX) > Length(CompBufferSSHC))
    then
      // if the result of SLZX is unsatisfying, use SSHC instead
      CompMethod := CompMethod_SSHC
    else
      CompMethod := CompMethod_SLZX
  end
  else if CompMethod = CompMethod_SSHC then
    ResultSSHC := Compress_SSHC(Buffer, CompBufferSSHC);

  // choose and return the right compressed buffer as determined above
  SetLength(Buffer, 0);
  if CompMethod = CompMethod_SLZX then
  begin
    SetLength(CompBufferSSHC, 0);
    Result := ResultSLZX;
    Buffer := CompBufferSLZX
  end
  else if CompMethod = CompMethod_SSHC then
  begin
    SetLength(CompBufferSLZX, 0);
    Result := ResultSSHC;
    Buffer := CompBufferSSHC
  end
end;

function AddFile(var SPRFile: file; const Dir: AnsiString;
  const SearchRec: TSearchRec; const DirTableIdx: Longint;
  var FileTable: TFileTable; var NameTable: TNameTable): Boolean;
var
  FileName,
  FileExt,
  FullName      : AnsiString;
  NameTableIdx,
  SuffixTableIdx,
  CompHeadSize  : Longint;
  InFile        : file;
  Buffer        : TBuffer;
  CRC           : TCRC;
begin
  Result := False;

  // add file name and extension to name table
  FileName := AnsiLowerCase(SearchRec.Name);
  FileExt := ExtractFileExt(FileName);
  NameTableIdx := AddToNameTable(Copy(FileName, 1,
    Length(FileName) - Length(FileExt)), NameTable);
  SuffixTableIdx := AddToNameTable(FileExt, NameTable);

  FullName := Format('%s%s', [Dir, SearchRec.Name]);

  // prepare file table entry
  SetLength(FileTable, Length(FileTable) + 1);
  with FileTable[High(FileTable)] do
  begin
    DirIdx     := DirTableIdx;
    NameIdx    := NameTableIdx;
    SuffixIdx  := SuffixTableIdx;
    CRC32      := 0;
    DateStamp  := GetDateStamp(FullName);
    Size       := SearchRec.Size;
    HeadOffset := -1;
    HeadSize   := 0;
    DataOffset := FilePos(SPRFile);
    DataSize   := SearchRec.Size
  end;

  // open input file
  if not OpenInFile(InFile, FullName) then
  begin
    WriteLn('  error opening input file');
    Exit
  end;

  // read data
  SetLength(Buffer, FileSize(InFile));
  {$I-}
  BlockRead(InFile, Buffer[0], Length(Buffer));
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  error reading file data');
    SetLength(Buffer, 0);
    CloseFile(InFile);
    Exit
  end;

  // compute and store CRC
  if not ParamRec.IsSave then
  begin
    CRC := TCRC.Create;
    CRC.Init;
    CRC.Update(Buffer[0], Length(Buffer));
    CRC.Final(FileTable[High(FileTable)].CRC32);
    CRC.Burn;
    CRC.Free
  end;

  // optionally compress the data
  CompHeadSize := Compress(Buffer, FileTable[High(FileTable)], NameTable);
  if CompHeadSize > 0 then
    with FileTable[High(FileTable)] do
    begin
      HeadOffset := FilePos(SPRFile);
      HeadSize   := CompHeadSize;
      DataOffset := FilePos(SPRFile) + CompHeadSize;
      DataSize   := Length(Buffer) - CompHeadSize
    end;

  // write data to archive
  {$I-}
  BlockWrite(SPRFile, Buffer[0], Length(Buffer));
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  error writing file data');
    SetLength(Buffer, 0);
    CloseFile(InFile);
    Exit
  end;

  SetLength(Buffer, 0);
  CloseFile(InFile);

  Result := True
end;

function AddDir(const SearchRec: TSearchRec; const DirTableIdx: Longint;
  var DirTable: TDirTable; var NameTable: TNameTable): Longint;
begin
  SetLength(DirTable, Length(DirTable) + 1);
  with DirTable[High(DirTable)] do
  begin
    // add directory name to name table
    NameIdx   := AddToNameTable(AnsiLowerCase(SearchRec.Name), NameTable);
    ParentIdx := DirTableIdx
  end;

  Result := High(DirTable)
end;

function ScanDir(var SPRFile: file; const Dir, RelDir: AnsiString;
  DirTableIdx: Longint; var DirTable: TDirTable; var FileTable: TFileTable;
  var NameTable: TNameTable): Boolean;
var
  SearchRec: TSearchRec;
  SubDirIdx: Longint;
begin
  Result := True;

  if FindFirst(Format('%s*', [Dir]), faAnyFile, SearchRec) = 0 then
    repeat
      with SearchRec do
        if (Attr and faDirectory) <> 0 then
        begin
          if (Name = '.') or (Name = '..') then
            Continue;

          SubDirIdx := AddDir(SearchRec, DirTableIdx, DirTable, NameTable);

          // recursion for subdirectories
          if not ScanDir(SPRFile, Format('%s%s\', [Dir, Name]),
            Format('%s%s\', [RelDir, Name]), SubDirIdx, DirTable, FileTable,
            NameTable) then
          begin
            Result := False;
            Break
          end
        end
        else
        begin
          WriteLn(Format('%s%s', [RelDir, Name]));

          if not AddFile(SPRFile, Dir, SearchRec, DirTableIdx, FileTable,
            NameTable) then
          begin
            Result := False;
            Break
          end
        end
    until FindNext(SearchRec) <> 0;

  SysUtils.FindClose(SearchRec)
end;

function CreateSPR: Boolean;
var
  BaseDir  : AnsiString;
  SPRFile  : file;
  SPRHeader: TSPRHeader;
  DirTable : TDirTable;
  FileTable: TFileTable;
  NameTable: TNameTable;
begin
  Result := True;

  // validate requested base directory
  BaseDir := IncludeTrailingPathDelimiter(ExpandFileName(ParamRec.DirSpec));
  if not DirectoryExists(BaseDir) then
  begin
    WriteLn(Format('"%s" is not a valid directory.', [BaseDir]));
    Exit
  end;

  // open output archive
  if not OpenSPRFile(SPRFile, ParamRec.FileSpec) then
  begin
    WriteLn(Format('Could not open "%s" for output.', [ParamRec.FileSpec]));
    Exit
  end;

  // prepare and write a temporary header
  SPRHeader.Seed := Random(255) + 1;
  Result := WriteHeader(SPRFile, SPRHeader);

  // search and store files starting at the root directory
  if Result then
    Result := ScanDir(SPRFile, BaseDir, '', $1effffff, DirTable, FileTable,
      NameTable);

  // store name table
  if Result then
    Result := WriteNameTable(SPRFile, SPRHeader, NameTable);

  // store directory table
  if Result then
    Result := WriteDirTable(SPRFile, SPRHeader, DirTable);

  // store file table
  if Result then
    Result := WriteFileTable(SPRFile, SPRHeader, FileTable);

  // write final header
  if Result then
    Result := WriteHeader(SPRFile, SPRHeader);

  SetLength(DirTable, 0);
  SetLength(FileTable, 0);
  SetLength(NameTable, 0);
  CloseFile(SPRFile)
end;

function ProcessParams: Boolean;
var
  i, j,
  Key   : Longint;
  KeyStr: AnsiString;
begin
  Result := False;

  if ParamCount = 0 then
    Exit;

  with ParamRec do
  begin
    FileSpec := '';
    DirSpec := '';
    KeyDomain := 0;
    IsDemo := False;
    IsSave := False;
    NoCompress := False;

    for i := 1 to ParamCount do
    begin
      if AnsiLowerCase(ParamStr(i)) = '/spcd' then
        IsDemo := True
      else if AnsiLowerCase(ParamStr(i)) = '/save' then
        IsSave := True
      else if AnsiLowerCase(ParamStr(i)) = '/nocompress' then
        NoCompress := True
      else if Copy(AnsiLowerCase(ParamStr(i)), 1, 5) = '/key:' then
      begin
        KeyStr := AnsiLowerCase(Copy(ParamStr(i), 6, Length(ParamStr(i)) - 5));

        for j := 0 to High(GameConfigs) do
          if AnsiLowerCase(GameConfigs[j].ID) = KeyStr then
            KeyDomain := GameConfigs[j].KeyDomain;

        if KeyDomain = 0 then
          if TryStrToInt(KeyStr, Key) then
            KeyDomain := Longword(Key)
          else
            Exit
      end
      else if ParamStr(i)[1] = '/' then
        Exit
      else if FileSpec = '' then
        FileSpec := ParamStr(i)
      else if DirSpec = '' then
        DirSpec := ParamStr(i)
      else
        Exit
    end;

    Result := (FileSpec <> '') and (DirSpec <> '') and (KeyDomain <> 0)
  end
end;

begin
  Randomize;

  if not ReadConfig then
    WriteLn(Format('Warning: %s', [LastError]));

  if ProcessParams then
    CreateSPR
  else
    WriteLn('Usage: MakSPR <SPRFile> <DataDir> /key:<key> [/spcd] [/save] [/n' +
      'ocompress]')
end.

