{ File    : $URL: file:///D:/SVN/svnDelphi/trunk/Geheimakte%20Tunguska/SPRExt/src/SPRExt.dpr $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2011-10-16 23:40:46 +0200 (So, 16 Okt 2011) $
  Revision: $Rev: 36 $

  Animation Arts SPR extractor.
}
program SPRExt;

{$APPTYPE CONSOLE}

{$R '..\res\version.res' '..\res\version.rc'}

uses
  KeyCfg in 'backend\KeyCfg.pas',
  SCRC in 'utils\SCRC.pas',
  SysUtils,
  Windows;

type
  TParamRec = record
    FileSpec   : AnsiString;
    Force      : Boolean;
    OverrideKey: Longword;
    KeyOnly    : Boolean;
  end;

  TSPRBlockDesc = packed record
    Num   : Longint;
    Offset: Longint;
    Size  : Longint
  end;

  TSPRHeader_Release = packed record
    ID         : array[0..3] of Char;
    Unknown1   : Longword;
    Seed       : Longword;
    Unknown2   : Longword;
    Unknown3   : Longword;
    Strings    : TSPRBlockDesc;
    Files      : TSPRBlockDesc;
    Directories: TSPRBlockDesc;
  end;

  TSPRHeader_Demo = packed record
    ID         : array[0..3] of Char;
    Unknown1   : Longword;
    Unknown2   : Longword;
    Seed       : Longword;
    Unknown3   : Longword;
    Unknown4   : Longword;
    Strings    : TSPRBlockDesc;
    Files      : TSPRBlockDesc;
    Unknown5   : Longword;
    Directories: TSPRBlockDesc;
  end;

  // unified header variant
  TSPRHeader = record
    ID         : array[0..3] of Char;
    Seed       : Longword;
    Strings    : TSPRBlockDesc;
    Files      : TSPRBlockDesc;
    Directories: TSPRBlockDesc;
   end;

  TSPRDate = TSystemTime;
  { format:
  TSPRDate = packed record
    Year     : Word;
    Month    : Word;
    DayOfWeek: Word;
    Day      : Word;
    Hour     : Word;
    Minute   : Word;
    Second   : Word;
    MSecond  : Word;
  end;}

  TFileEntry = packed record
    DirIdx    : Longint;
    NameIdx   : Longint;
    SuffixIdx : Longint;
    CRC32     : Longword;
    DateStamp : TSPRDate;
    Size      : Longint;
    HeadOffset: Longint;
    HeadSize  : Longint;
    DataOffset: Longint;
    DataSize  : Longint;
  end;

  TDirEntry = packed record
    NameIdx  : Longint;
    ParentIdx: Longint;
  end;

  TBuffer    = array of Byte;
  TNameTable = array of string;
  TFileTable = array of TFileEntry;
  TDirTable  = array of string;

var
  ParamRec   : TParamRec;
  DetectedKey: Longword;

function OpenSPRFile(var SPRFile: file; SPRName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(SPRFile, SPRName);

  {$I-}
  Reset(SPRFile, 1);
  {$I+}
  OpenSPRFile := (IOResult = 0)
end;

function ReadHeader(var SPRFile: file; out SPRHeader: TSPRHeader): Boolean;
var
  ID        : array[0..3] of Char;
  BytesRead : Longint;
  SPRRelease: TSPRHeader_Release;
  SPRDemo   : TSPRHeader_Demo;
begin
  BlockRead(SPRFile, ID, Length(ID), BytesRead);
  Result := (Length(ID) = BytesRead);
  if not Result then
    Exit;

  if ID = 'SPCR' then
  begin
    Seek(SPRFile, 0);
    BlockRead(SPRFile, SPRRelease, SizeOf(SPRRelease), BytesRead);
    Result := (SizeOf(SPRRelease) = BytesRead);
    if not Result then
      Exit;

    with SPRHeader do
    begin
      Move(SPRRelease.ID, ID, SizeOf(ID));
      Move(SPRRelease.Seed, Seed, SizeOf(Seed));
      Move(SPRRelease.Strings, Strings, SizeOf(Strings));
      Move(SPRRelease.Files, Files, SizeOf(Files));
      Move(SPRRelease.Directories, Directories, SizeOf(Directories))
    end
  end
  else if ID = 'SPCD' then
  begin
    Seek(SPRFile, 0);
    BlockRead(SPRFile, SPRDemo, SizeOf(SPRDemo), BytesRead);
    Result := (SizeOf(SPRDemo) = BytesRead);
    if not Result then
      Exit;

    with SPRHeader do
    begin
      Move(SPRDemo.ID, ID, SizeOf(ID));
      Move(SPRDemo.Seed, Seed, SizeOf(Seed));
      Move(SPRDemo.Strings, Strings, SizeOf(Strings));
      Move(SPRDemo.Files, Files, SizeOf(Files));
      Move(SPRDemo.Directories, Directories, SizeOf(Directories))
    end
  end
end;

procedure DecryptBuffer(const Buffer: TBuffer; const SPRHeader: TSPRHeader);
var
  i        : Longint;
  KeyDomain,
  Key      : Longword;
begin
  // 0 is an invalid key domain, since no mod operation is possible
  if DetectedKey = $00000000 then
    Exit;

  KeyDomain := DetectedKey * SPRHeader.Seed;
  Key := ((SPRHeader.Seed shl 8) or ((not SPRHeader.Seed) and $000000ff));

  for i := 0 to High(Buffer) do
  begin
    Key := (Key * Key) mod KeyDomain;
    Buffer[i] := Buffer[i] xor Key;
    Key := Key + Buffer[i] + 1
  end
end;

procedure Decompress_SLZX(const CompHead: TBuffer; const InBuf: TBuffer;
  out OutBuf: TBuffer; out UncompSize: Longint);
const
  NBits = 11;
type
  TSLZXHeader = packed record
    ID        : array[0..3] of Char;
    UncompSize: Longword;
  end;
var
  Header   : TSLZXHeader;
  N,
  OfsLen,
  Len      : Word;
  Offset   : Longint;
  Flags    : Byte;
  i, j,
  InBufPos,
  OutBufPos: Longint;
begin
  SetLength(OutBuf, 0);
  UncompSize := -1;

  if Length(CompHead) < SizeOf(Header) then
    Exit;

  Move(CompHead[0], Header, SizeOf(Header));

  if Header.ID <> 'SLZX' then
    Exit; // shouldn't occur

  // prepare output
  SetLength(OutBuf, Header.UncompSize);
  UncompSize := Header.UncompSize;

  // skip additional uncompressed size
  InBufPos := 4;
  OutBufPos := 0;

  // initialize LZSS parameters for this block
  N := 1 shl NBits - 1;

  // start extracting block contents
  while InBufPos < Length(InBuf) do
  begin
    // read bit field
    Flags := InBuf[InBufPos];
    Inc(InBufPos);

    // process bit field
    for i := 1 to 8 do
    begin
      if InBufPos <= High(InBuf) then
      begin
        // check for buffer window reference
        if (Flags and 1 = 1) then
        begin
          if InBufPos + 1 > High(InBuf) then
          begin
            SetLength(OutBuf, 0);
            Exit
          end;

          // read offset and length
          Move(InBuf[InBufPos], OfsLen, 2);
          Inc(InBufPos, 2);

          Len := OfsLen shr NBits;

          Offset := OfsLen and N;
          if OutBufPos > N + 1 then
            Inc(Offset, OutBufPos - (N + 1));

          if (OutBufPos + Len - 1 > High(OutBuf)) or
            (Offset + Len - 1 > High(OutBuf)) then
          begin
            SetLength(OutBuf, 0);
            Exit
          end;

          // peek into buffer
          for j := 0 to Len - 1 do
          begin
            OutBuf[OutBufPos] := OutBuf[Offset + j];
            Inc(OutBufPos);
          end;

          // emit one additional character if possible
          if (InBufPos <= High(InBuf)) and (OutBufPos <= High(OutBuf)) then
          begin
            OutBuf[OutBufPos] := InBuf[InBufPos];
            Inc(InBufPos);
            Inc(OutBufPos)
          end
        end
        else
        // copy literally
        begin
          if (InBufPos > High(InBuf)) or (OutBufPos > High(OutBuf)) then
            Exit;

          OutBuf[OutBufPos] := InBuf[InBufPos];
          Inc(InBufPos);
          Inc(OutBufPos)
        end;

        Flags := Flags shr 1
      end
    end
  end
end;

procedure Decompress_SSHC(const CompHead: TBuffer; const InBuf: TBuffer;
  out OutBuf: TBuffer; out UncompSize: Longint);
const
  Null: Word = $0fff;
type
  TSSHCHeader = packed record
    ID        : array[0..3] of Char;
    UncompSize: Longword;
    CompSize  : Longword;
    NumNodes  : Longword;
    NumBits   : Longword;
  end;
  THuffNode = record
    Parent    : Word;
    LeftChild : Word;
    RightChild: Word;
    Value     : Word;
  end;
  PHuffNode = ^THuffNode;
  THuffTree = array of PHuffNode;
var
  i,
  InBufPos,
  OutBufPos: Longint;
  Header   : TSSHCHeader;
  HuffNode : PHuffNode;
  HuffTree : THuffTree;
  RootID,
  NodeID   : Word;
  CurBit,
  CurByte  : Byte;
begin
  SetLength(OutBuf, 0);
  UncompSize := -1;

  if Length(CompHead) < SizeOf(Header) then
    Exit;

  Move(CompHead[0], Header, SizeOf(Header));

  if Header.ID <> 'SSHC' then
    Exit; // shouldn't occur

  InBufPos := 0;
  OutBufPos := 0;

  // prepare Huffman tree (way too much memory used here, but who cares ...)
  SetLength(HuffTree, 65536);
  for i := 0 to 65535 do
    HuffTree[i] := nil;

  // build Huffman tree
  RootID := Null;
  for i := 1 to Header.NumNodes do
  begin
    if InBufPos + SizeOf(HuffNode^) > High(InBuf) + 1 then
      Break;

    New(HuffNode);
    Move(InBuf[InBufPos], HuffNode^, SizeOf(HuffNode^));
    Inc(InBufPos, SizeOf(HuffNode^));

    HuffTree[HuffNode^.Value] := HuffNode;

    // check for root node definition
    if HuffNode^.Parent = Null then
      RootID := HuffNode^.Value
  end;

  // prepare output
  SetLength(OutBuf, Header.UncompSize);
  UncompSize := Header.UncompSize;

  // decompress data block
  InBufPos := InBufPos shl 3;
  CurByte := 0;
  NodeID := RootID;
  for i := 1 to Header.NumBits do
  begin
    if InBufPos and $00000007 = 0 then
    begin
      if InBufPos shr 3 > High(InBuf) then
        Break;
      CurByte := InBuf[InBufPos shr 3]
    end
    else
      CurByte := CurByte shr 1;

    // get next bit from data stream
    CurBit := CurByte and $01;
    Inc(InBufPos);

    // get next literal by traversing the Huffman tree
    if CurBit = 0 then
      NodeID := HuffTree[NodeID]^.LeftChild
    else
      NodeID := HuffTree[NodeID]^.RightChild;

    HuffNode := HuffTree[NodeID];

    if (HuffNode^.LeftChild = Null) and (HuffNode^.RightChild = Null) then
    begin
      // leaf node found: output literal
      if OutBufPos <= High(OutBuf) then
      begin
        OutBuf[OutBufPos] := HuffNode^.Value and $00ff;
        Inc(OutBufPos);
        // if the output buffer is full, end the decompression
        if OutBufPos > High(OutBuf) then
          Break;
      end;

      // reset to tree root
      NodeID := RootID
    end
    else
      // inner node found: branch to next node
      NodeID := HuffNode^.Value
  end;

  for i := 0 to High(HuffTree) do
    if Assigned(HuffTree[i]) then
      Dispose(HuffTree[i]);
  SetLength(HuffTree, 0)
end;

procedure Decompress(const CompHead: TBuffer; var Buffer: TBuffer);
var
  CompID    : array[0..3] of Char;
  UncompSize: Longint;
  UncompBuf : TBuffer;
begin
  if Length(CompHead) < Length(CompID) then
    Exit;

  Move(CompHead[0], CompID, Length(CompID));

  if CompID = 'SLZX' then
    Decompress_SLZX(CompHead, Buffer, UncompBuf, UncompSize)
  else if CompID = 'SSHC' then
    Decompress_SSHC(CompHead, Buffer, UncompBuf, UncompSize)
  else
  begin
    WriteLn(Format('    unknown compression "%s"' , [string(CompID)]));
    Exit
  end;

  if Length(UncompBuf) <> UncompSize then
    WriteLn('    decompression error')
  else
  begin
    SetLength(Buffer, 0);
    Buffer := UncompBuf
  end
end;

function DetectGameKey(const Buffer: TBuffer;
  const SPRHeader: TSPRHeader): Boolean;

  // The following test routine has been derived from the general decryption
  // procedure to keep memory copy operations to a minimum. This looks quite
  // awful, but improves the performance of the brute force search noticeably.
  function DecryptAndCheck(const InBuf, OutBuf: TBuffer;
    const SPRHeader: TSPRHeader): Boolean;
  var
    i,
    Boundary : Longint;
    KeyDomain,
    Key      : Longword;
    Len      : Word;
  begin
    Result := False;

    // 0 is an invalid key domain, since no mod operation is possible
    if DetectedKey = $00000000 then
      Exit;

    KeyDomain := DetectedKey * SPRHeader.Seed;
    Key := ((SPRHeader.Seed shl 8) or ((not SPRHeader.Seed) and $000000ff));

    // this is the the value the accumulated string length should reach
    Boundary := Length(InBuf) - SPRHeader.Strings.Num * 2;

    for i := 0 to High(InBuf) do
    begin
      Key := (Key * Key) mod KeyDomain;
      OutBuf[i] := InBuf[i] xor Key;
      Key := Key + OutBuf[i] + 1;

      // if another word value has been decrypted, update the remaining boundary
      if ((i and 1) = 1) and ((i - 1) shr 1 < SPRHeader.Strings.Num) then
      begin
        Move(OutBuf[i - 1], Len, 2);
        Dec(Boundary, Len);

        // if the boundary has been violated, there's no need to test further
        if Boundary < 0 then
          Exit
      end
    end;

    // the test is successful only if the accumulated string length matches
    // exactly
    Result := (Boundary = 0)
  end;

var
  GameIdx: Longint;
  TempKey: Longword;
  TmpBuf : TBuffer;
begin
  Result := False;

  // any valid string buffer should contain at least one string length;
  if Length(Buffer) < 2 then
    Exit;

  // set up the temporary target buffer for the check routine
  SetLength(TmpBuf, Length(Buffer));

  // if an override key has been specified, then that one takes precedence
  if ParamRec.OverrideKey > $00000000 then
  begin
    WriteLn('  Info: Using override key 0x' + AnsiLowerCase(IntToHex(
      ParamRec.OverrideKey, 8)));
    DetectedKey := ParamRec.OverrideKey;

    if DecryptAndCheck(Buffer, TmpBuf, SPRHeader) then
      Result := True
    else
      WriteLn('  Info: Override key is invalid.')
  end;

  // Try to detect the game data encryption using known key domains: The various
  // Deep Silver games using SPR files use the same file format, but different
  // key domains.
  if not Result then
    for GameIdx := 0 to High(GameConfigs) do
    begin
      DetectedKey := GameConfigs[GameIdx].KeyDomain;

      if DecryptAndCheck(Buffer, TmpBuf, SPRHeader) then
      begin
        WriteLn('  Info: Assuming ' + GameConfigs[GameIdx].Name +
          ' data file.');
        Result := True;
        Break
      end
    end;

  // If no known key matches and brute force is allowed, let's try to find the
  // key now. The key space can be expected to be exhausted within 5 to 10
  // minutes on a contemporary machine (depending on the exact buffer given).
  // Since the currently observed game keys are in a relatively low range, the
  // algorithm usually takes only a fraction of a second to complete.
  if not Result and ParamRec.Force then
  begin
    // if an override key has been specified, take off from there ...
    if ParamRec.OverrideKey > $00000000 then
    begin
      WriteLn('  Info: Trying to find unknown key, starting at override key.');
      TempKey := ParamRec.OverrideKey;
    end
    // ... else start from the beginning
    else
    begin
      WriteLn('  Info: Trying to find unknown key.');
      TempKey := $00000001;
    end;

    while Longword(TempKey) <> $00000000 do
    begin
      if TempKey and $00ffffff = 0 then
        WriteLn('  Info: Currently trying key 0x' + AnsiLowerCase(IntToHex(
          TempKey, 8)));

      DetectedKey := TempKey;

      // test current key
      if DecryptAndCheck(Buffer, TmpBuf, SPRHeader) then
      begin
        WriteLn('  Info: Found key 0x' + AnsiLowerCase(IntToHex(DetectedKey
          , 8)));
        Result := True;
        Break
      end;

      Inc(TempKey, 1)
    end;
  end;

  if not Result then
    WriteLn('  Info: No matching key found.');

  SetLength(TmpBuf, 0)
end;

function ReadStructures(var SPRFile: file; const SPRHeader: TSPRHeader;
  var NameTable: TNameTable; var FileTable: TFileTable;
  var DirTable: TDirTable): Boolean;
var
  i,
  BufIdx  : Longint;
  Len     : Word;
  DirEntry: TDirEntry;
  Buffer  : TBuffer;
begin
  // load string table data
  if SPRHeader.Strings.Offset > 0 then
  begin
    SetLength(Buffer, SPRHeader.Strings.Size);
    {$I-}
    Seek(SPRFile, SPRHeader.Strings.Offset);
    BlockRead(SPRFile, Buffer[0], Length(Buffer));
    {$I+}
    Result := (IOResult = 0);
    if not Result then
    begin
      SetLength(Buffer, 0);
      Exit
    end;

    // get the game key
    Result := DetectGameKey(Buffer, SPRHeader);
    if not Result then
    begin
      WriteLn('  error finding decryption key');
      Exit
    end;

    // decrypt string table data
    DecryptBuffer(Buffer, SPRHeader);

    // convert to indexed name table
    SetLength(NameTable, SPRHeader.Strings.Num);
    for i := 0 to High(NameTable) do
    begin
      Move(Buffer[i * 2], Len, 2);
      SetLength(NameTable[i], Len)
    end;

    BufIdx := SPRHeader.Strings.Num * 2;
    for i := 0 to High(NameTable) do
    begin
      Move(Buffer[BufIdx], NameTable[i][1], Length(NameTable[i]));
      Inc(BufIdx, Length(NameTable[i]))
    end
  end
  else
    SetLength(NameTable, 0);

  // load directory table data
  if SPRHeader.Directories.Offset > 0 then
  begin
    SetLength(Buffer, SPRHeader.Directories.Size);
    {$I-}
    Seek(SPRFile, SPRHeader.Directories.Offset);
    BlockRead(SPRFile, Buffer[0], Length(Buffer));
    {$I+}
    Result := (IOResult = 0);
    if not Result then
    begin
      SetLength(Buffer, 0);
      Finalize(NameTable);
      SetLength(NameTable, 0);
      Exit
    end;

    // decrypt directory table data
    DecryptBuffer(Buffer, SPRHeader);

    // construct full directory names
    SetLength(DirTable, SPRHeader.Directories.Num);
    for i := 0 to High(DirTable) do
    begin
      Move(Buffer[i * SizeOf(DirEntry)], DirEntry, SizeOf(DirEntry));

      with DirEntry do
        // check for root children
        if ParentIdx <> $1effffff then
          DirTable[i] := DirTable[ParentIdx] + NameTable[NameIdx] + '\'
        else
          DirTable[i] := NameTable[NameIdx] + '\'
    end
  end
  else
    SetLength(DirTable, 0);

  // load file table data
  if SPRHeader.Files.Offset > 0 then
  begin
    SetLength(Buffer, SPRHeader.Files.Size);
    {$I-}
    Seek(SPRFile, SPRHeader.Files.Offset);
    BlockRead(SPRFile, Buffer[0], Length(Buffer));
    {$I+}
    Result := (IOResult = 0);
    if not Result then
    begin
      SetLength(Buffer, 0);
      Finalize(NameTable);
      SetLength(NameTable, 0);
      Finalize(DirTable);
      SetLength(DirTable, 0);
      Exit
    end;

    // decrypt file table data
    DecryptBuffer(Buffer, SPRHeader);

    // copy to structured file table
    SetLength(FileTable, SPRHeader.Files.Num);
    Move(Buffer[0], FileTable[0], Length(Buffer))
  end
  else
    SetLength(FileTable, 0);

  SetLength(Buffer, 0);
  
  Result := True
end;

procedure SetDateStamp(const FileName: string; DateStamp: TSystemTime);
var
  FileTime: TFileTime;
  Handle  : Longword;
begin
  SystemTimeToFileTime(DateStamp, FileTime);
  LocalFileTimeToFileTime(FileTime, FileTime);

  Handle := FileOpen(FileName, fmOpenWrite);
  if Handle = INVALID_HANDLE_VALUE then
    Exit;

  SetFileTime(Handle, @FileTime, @FileTime, @FileTime);
  
  FileClose(Handle)
end;

procedure ExtractFiles(var SPRFile: file; const SPRHeader: TSPRHeader);
var
  i        : Longint;
  CompHead,
  Buffer   : TBuffer;
  NameTable: TNameTable;
  FileTable: TFileTable;
  DirTable : TDirTable;
  FileName : string;
  CRC      : TCRC;
  CRCVal   : Longword;
  OutFile  : file;
begin
  if not ReadStructures(SPRFile, SPRHeader, NameTable, FileTable, DirTable) then
  begin
    WriteLn('  error reading structures');
    Exit
  end
  else if ParamRec.KeyOnly then
    Exit;

  CRC := TCRC.Create;

  // process files
  for i := 0 to High(FileTable) do
  begin
    with FileTable[i] do
    begin
      FileName := '';

      if (DirIdx >= 0) and (DirIdx <= High(DirTable)) then
        FileName := FileName + DirTable[DirIdx];
      if (NameIdx >= 0) and (NameIdx <= High(NameTable)) then
        FileName := FileName + NameTable[NameIdx];
      if (SuffixIdx >= 0) and (SuffixIdx <= High(NameTable)) then
        FileName := FileName + NameTable[SuffixIdx];

      WriteLn('  ', FileName);

      // read compression header if present
      SetLength(CompHead, HeadSize);
      if HeadOffset <> -1 then
      begin
        {$I-}
        Seek(SPRFile, HeadOffset);
        BlockRead(SPRFile, CompHead[0], Length(CompHead));
        {$I+}
        if IOResult <> 0 then
        begin
          WriteLn('    file read error');
          Continue
        end
      end;

      // read file data
      {$I-}
      SetLength(Buffer, DataSize);
      Seek(SPRFile, DataOffset);
      BlockRead(SPRFile, Buffer[0], Length(Buffer));
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    file read error');
        Continue
      end;

      // decompress if necessary
      if HeadOffset <> -1 then
        Decompress(CompHead, Buffer);

      // create directory
      if (DirIdx >= 0) and (DirIdx <= High(DirTable)) then
        ForceDirectories(ExpandFileName(DirTable[DirIdx]));

      // create output file
      AssignFile(OutFile, FileName);
      {$I-}
      Rewrite(OutFile, 1);
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    file creation error');
        Continue
      end;

      // extract file
      {$I-}
      BlockWrite(OutFile, Buffer[0], Length(Buffer));
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    write error');
        CloseFile(OutFile);
        Continue
      end;

      CloseFile(OutFile);

      // set file date
      with DateStamp do
        SetDateStamp(FileName, DateStamp);

      // check CRC
      if CRC32 <> 0 then
      begin
        CRC.Init;
        CRC.Update(Buffer[0], Length(Buffer));
        CRC.Final(CRCVal);

        if CRCVal <> CRC32 then
          WriteLn('    CRC error');

        CRC.Burn
      end
    end
  end;

  // clean up
  CRC.Free;

  SetLength(Buffer, 0);
  SetLength(CompHead, 0);
  Finalize(NameTable);
  SetLength(NameTable, 0);
  Finalize(FileTable);
  SetLength(FileTable, 0);
  Finalize(DirTable);
  SetLength(DirTable, 0)
end;

procedure ScanFiles;
var
  FileName,
  Dir,
  SPRName  : string;
  SPRFile  : file;
  SPRHeader: TSPRHeader;
  SearchRec: TSearchRec;
begin
  // get the directory information
  Dir := ExtractFilePath(ParamRec.FileSpec);

  // search for all files matching the file specification
  if FindFirst(ParamRec.FileSpec, faAnyFile - faDirectory,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      SPRName := Dir + FileName;

      WriteLn(SPRName);

      if OpenSPRFile(SPRFile, SPRName) then
      begin
        if ReadHeader(SPRFile, SPRHeader) then
          ExtractFiles(SPRFile, SPRHeader)
        else
          WriteLn('  not a valid SPR file')
      end
      else
        WriteLn('  not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(Format('No matches for "%s"', [ParamRec.FileSpec]));

  SysUtils.FindClose(SearchRec)
end;

function ProcessParams: Boolean;
var
  i,
  Key   : Longint;
  KeyStr: AnsiString;
begin
  Result := False;

  if ParamCount = 0 then
    Exit;

  with ParamRec do
  begin
    FileSpec := '';
    Force := False;
    OverrideKey := 0;
    KeyOnly := False;

    for i := 1 to ParamCount do
    begin
      if AnsiLowerCase(ParamStr(i)) = '/force' then
        Force := True
      else if Copy(AnsiLowerCase(ParamStr(i)), 1, 13) = '/overridekey:' then
      begin
        KeyStr := Copy(ParamStr(i), 14, Length(ParamStr(i)) - 13);
        if TryStrToInt(KeyStr, Key) then
          OverrideKey := Longword(Key)
        else
          Exit
      end
      else if Copy(AnsiLowerCase(ParamStr(i)), 1, 8) = '/keyonly' then
        KeyOnly := True
      else if ParamStr(i)[1] = '/' then
        Exit
      else if FileSpec <> '' then
        Exit
      else
        FileSpec := ParamStr(i)
    end;

    Result := (FileSpec <> '')
  end
end;

begin
  if ProcessParams then
  begin
    if not ReadConfig then
      WriteLn(Format('Warning: %s', [LastError]));
    ScanFiles
  end
  else
    WriteLn('Usage: SPRExt <files> [/force] [/overrideKey:<key>] [/keyonly]')
end.

